use async_channel::{Receiver, Sender};
use tokio::runtime::Runtime;
use xmpp::{BareJid, ClientBuilder, ClientFeature};

use std::path::PathBuf;
use std::sync::OnceLock;

#[derive(Clone, Debug)]
pub enum XMPPCommand {
    SendPM(BareJid, String),
}

#[derive(Clone, Debug)]
pub enum XMPPEvent {
    Avatar(BareJid, String),
    Contact(BareJid, Option<String>),
}

pub fn tokio_runtime() -> &'static Runtime {
    static RUNTIME: OnceLock<Runtime> = OnceLock::new();
    RUNTIME.get_or_init(|| Runtime::new().expect("Setting up tokio runtime needs to succeed."))
}

pub async fn handle_xmpp_event(event: xmpp::Event) -> Option<XMPPEvent> {
    match event {
        xmpp::Event::AvatarRetrieved(jid, hash) => {
            // need update latest symlink?
            // xmpp-rs stores avatar in data/JID/HASH... we create data/JID/latest symlink points to relative HASH

            let mut latest_link = PathBuf::from(&hash);
            let hash_name = latest_link
                .file_name()
                .unwrap()
                .to_str()
                .unwrap()
                .to_string();
            latest_link.set_file_name("latest");

            if let Ok(metadata) = tokio::fs::metadata(&latest_link).await {
                if metadata.is_symlink() {
                    if let Ok(previous_latest) = tokio::fs::read_link(&latest_link).await {
                        if previous_latest == PathBuf::from(&hash_name) {
                            // We already have latest symlink
                            log::info!("avatar Already have latest link {jid} to hash {hash}");
                            return None;
                        }
                    }
                }
            }

            // Setting new latest symlink
            let _ = tokio::fs::symlink(hash_name, latest_link).await;

            if jid.is_bare() {
                Some(XMPPEvent::Avatar(jid.to_bare(), hash))
            } else {
                // no avatar for FullJid (yet)
                None
            }
        }
        xmpp::Event::ContactAdded(jid) => {
            let avatar_jid = jid.jid.clone();
            if let Ok(hash_name) =
                tokio::fs::read_link(&format!("data/{}/latest", &avatar_jid)).await
            {
                let avatar = format!("data/{}/{}", &avatar_jid, hash_name.display());
                log::info!("Found existing avatar for contact: {avatar_jid}");
                Some(XMPPEvent::Contact(avatar_jid, Some(avatar)))
            } else {
                Some(XMPPEvent::Contact(avatar_jid, None))
            }
        }
        xmpp::Event::ChatMessage(_id, _from, _body, _time) => {
            // TODO: Insert message into tab history
            None
        }
        _ => None,
    }
}

pub(crate) fn client(jid: &str, password: &str) -> (Receiver<XMPPEvent>, Sender<XMPPCommand>) {
    let (event_sender, event_receiver) = async_channel::bounded::<XMPPEvent>(1);
    let (cmd_sender, cmd_receiver) = async_channel::bounded::<XMPPCommand>(1);

    let jid = jid.to_string();
    let password = password.to_string();

    tokio_runtime().spawn(async move {
        let mut client = ClientBuilder::new(BareJid::new(&jid).unwrap(), &password)
            .set_default_nick("xmpp-client")
            .enable_feature(ClientFeature::Avatars)
            .enable_feature(ClientFeature::ContactList)
            .build();

        loop {
            tokio::select! {
                Some(events) = client.wait_for_events() => {
                    for event in events {
                        if let Some(parsed_event) = handle_xmpp_event(event).await {
                            let _ = event_sender.send(parsed_event).await;
                        }
                    }
                }

                command = cmd_receiver.recv() => {
                    match command.unwrap() {
                        XMPPCommand::SendPM(jid, content) => {
                            client.send_message(jid.into(), xmpp::parsers::message::MessageType::Chat, "en", &content).await;
                        },
                    }
                }
            }
        }
    });

    (event_receiver, cmd_sender)
}
